###
### "THE BEER-WARE LICENSE":
### Alberto Ramos wrote this file. As long as you retain this 
### notice you can do whatever you want with this stuff. If we meet some 
### day, and you think this stuff is worth it, you can buy me a beer in 
### return. <alberto.ramos@cern.ch>
###
### file:    ana.jl
### created: Tue Nov  9 17:28:04 2021
###                               

using Printf, Plots, BDIO, ADerrors, ArgParse


function parse_commandline()
    s = ArgParseSettings()

    @add_arg_table s begin
        "-i"
        help = "input file"
        required = true
        arg_type = String
    end

    @add_arg_table s begin
        "-o"
        help = "output directory"
        required = true
        arg_type = String
    end

    @add_arg_table s begin
        "-c"
        help = "Value of t^2<E(t)> to determine t0,w0 scales"
        required = false
        arg_type = Float64
        default = 0.3
    end
    
    return parse_args(s)
end

function get_msm_info(fb)

    nthm = 0
    nmsm = 0
    nflw = 0
    flw_ns = 0
    eps = 0.0
    BDIO_seek!(fb, 0)
    while BDIO_seek!(fb)
        if BDIO_get_uinfo(fb) == 2
            nthm = round(Int64, BDIO_get_len(fb)/8)
        end
        if BDIO_get_uinfo(fb) == 3
            nmsm = round(Int64, BDIO_get_len(fb)/8)
        end
        if BDIO_get_uinfo(fb) == 12
            nflw = round(Int64, BDIO_get_len(fb)/8-2)
            foo = zeros(2)
            BDIO_read(fb, foo)
            flw_ns = round(Int64, foo[2])
            eps = foo[1]
        end
    end

    return nthm, nmsm, round(Int64, nflw/(flw_ns+1)), eps, flw_ns
end

function read_record(dt::Vector, fb, uinfo)

    BDIO_seek!(fb, 0)
    while BDIO_seek!(fb)
        if BDIO_get_uinfo(fb) == uinfo
            BDIO_read(fb, dt)
            return true
        end
    end

    return false
end

function read_record(dt::Array{Float64, 2}, fb, uinfo)

    BDIO_seek!(fb, 0)
    while BDIO_seek!(fb)
        if BDIO_get_uinfo(fb) == uinfo
            foo = Vector{Float64}(undef, size(dt,1))
            for i in 1:size(dt,2)
                BDIO_read(fb, foo)
                for j in 1:length(foo)
                    dt[j,i] = foo[j]
                end
            end
            return true
        end
    end

    return false
end

function ana_record(fb, nmsm, uinfo)

    dt = Vector{Float64}(undef, nmsm)
    read_record(dt, fb, uinfo)
    if uinfo == 3
        obs = uwreal(exp.(-dt), "Ensemble")
    else
        obs = uwreal(dt, "Ensemble")
    end
    uwerr(obs)
    
    return obs
end

function ana_flow_record(fb, nflw, flw_ns, eps, uinfo, tsqet::Union{Nothing, Float64} = nothing)

    dt = Array{Float64, 2}(undef, flw_ns+1, nflw)
    if !read_record(dt, fb, uinfo)
        if tsqet == nothing
            return nothing
        else
            return nothing, nothing, nothing, nothing
        end
    end
    
    for i in 1:size(dt, 1)
        t = (i-1)*eps
        for j in 1:size(dt,2)
            dt[i,j] = dt[i,j]*t^2
        end
    end

    obs = Vector{uwreal}(undef, flw_ns+1)
    for i in 1:flw_ns+1
        obs[i] = uwreal(dt[i,:], "Flow")
    end
    uwerr.(obs)

    if tsqet == nothing
        return obs
    else
        # Now determine t0 scale
        id = 0
        d = 100.0
        for i in 1:length(obs)
            if abs(value(obs[i])-tsqet) < d
                d = abs(value(obs[i])-tsqet)
                id = i
            end
        end
        ta = (id-2)*eps
        tb = (id-1)*eps
        B = (obs[id-1]-obs[id])/(ta-tb)
        A = obs[id-1] - B*ta

        t0 = (tsqet-A)/B

        # w0 scale
        dobs = Vector{uwreal}(undef, length(obs)-2)
        for i in 2:length(obs)-1
            dobs[i-1] = i*(obs[i+1]-obs[i-1])/2
        end

        id = 0
        d = 100.0
        for i in 1:length(dobs)
            if abs(value(dobs[i])-tsqet) < d
                d = abs(value(dobs[i])-tsqet)
                id = i
            end
        end
        ta = (id-1)*eps
        tb = (id)*eps
        B = (obs[id-1]-obs[id])/(ta-tb)
        A = obs[id-1] - B*ta

        w0 = sqrt((tsqet-A)/B)
        
        
        return obs, t0, w0, id
    end


    return nothing
end

function plot_et(plq, clv, eps, tsqet)

    t = [(i-1)*eps for i in 1:length(plq)]
    plot(xlabel="t", ylabel="t^2<E(t)>",legend=:topleft)
    plot!(t, value.(plq), yerr=err.(plq), label="Plaquette", seriestype=:scatter)
    plot!(t, value.(clv), yerr=err.(clv), label="Clover", seriestype=:scatter)
    pl = hline!([tsqet],label="",width=2,color=:black)

    return pl
end


function make_plot(obs)

    ens = ensembles(obs)
    l = grid(2,1, heights=[0.3,0.7])
    pl1 = plot(mchist(obs, ens[1]) .+ value(obs), xlabel="MC time",
               ylabel="", label="")
    r  =  rho(obs, ens[1])
    dr = drho(obs, ens[1])

    w = max(20, 2*window(obs, ens[1]))
    pl2 = plot(r[1:w], yerr=dr[1:w], xlabel="W", ylabel="AC function",
               seriestype=:scatter, label="")
    pl2 = vline!([window(obs, ens[1])], label="")
    
    pl = plot(pl1, pl2, layout=l)

    return pl
end


parsed_args = parse_commandline()
fname   = parsed_args["i"]
dout    = parsed_args["o"]
tsqet   = parsed_args["c"]

fb = BDIO_open(fname, "r")
nthm, nmsm, nflw, eps, flw_ns = get_msm_info(fb)

fout = open(dout*"/results.txt", "w")
println(fout, "\n## Results of analysis of file: "*fname)
println(fout, "  - Number of thermalization steps: ", nthm)
println(fout, "  - Number of measurements:         ", nmsm)
println(fout, "  - flow measurements for times:    ", 0.0, " - ", flw_ns*eps)
println(fout, "## ")

println(fout, " # Observable <exp(-dH)>")
obs = ana_record(fb, nmsm, 3)
println(fout, "  - value:  ", obs)
println(fout, "  - tauint: ", taui(obs, "Ensemble"), " +/- ", dtaui(obs, "Ensemble"))
pl = make_plot(obs)
savefig(pl, dout*"/IamOne.pdf")

println(fout, " # Observable <P>")
obs = ana_record(fb, nmsm, 5)
println(fout, "  - value:  ", obs)
println(fout, "  - tauint: ", taui(obs, "Ensemble"), " +/- ", dtaui(obs, "Ensemble"))
pl = make_plot(obs)
savefig(pl, dout*"/plaq.pdf")

plq_t0, t0, w0, id = ana_flow_record(fb, nflw, flw_ns, eps, 12, tsqet)
if t0 == nothing
    println(fout, " # No flow observables on file")
else
    println(fout, " # Observable t0 plaquette")
    uwerr(t0)
    println(fout, "  - value:  ", t0)
    println(fout, "  - tauint: ", taui(t0, "Flow"), " +/- ", dtaui(t0, "Flow"))
    pl = make_plot(t0)
    savefig(pl, dout*"/t0_pl.pdf")

    println(fout, " # Observable w0 plaquette")
    uwerr(w0)
    println(fout, "  - value:  ", w0)
    println(fout, "  - tauint: ", taui(w0, "Flow"), " +/- ", dtaui(w0, "Flow"))
    pl = make_plot(w0)
    savefig(pl, dout*"/w0_pl.pdf")
end
    

clv_t0, t0, w0, id = ana_flow_record(fb, nflw, flw_ns, eps, 13, tsqet)
if t0 != nothing
    println(fout, " # Observable t0 clover")
    uwerr(t0)
    println(fout, "  - value:  ", t0)
    println(fout, "  - tauint: ", taui(t0, "Flow"), " +/- ", dtaui(t0, "Flow"))
    pl = make_plot(t0)
    savefig(pl, dout*"/t0_cl.pdf")
    
    println(fout, " # Observable w0 clover")
    uwerr(w0)
    println(fout, "  - value:  ", w0)
    println(fout, "  - tauint: ", taui(w0, "Flow"), " +/- ", dtaui(w0, "Flow"))
    pl = make_plot(w0)
    savefig(pl, dout*"/w0_cl.pdf")
    
    pl = plot_et(plq_t0, clv_t0, eps, tsqet)
    savefig(pl, dout*"/tsq_et.pdf")
    
    obs = ana_flow_record(fb, nflw, flw_ns, eps, 14, nothing)
    qtop = obs[id]
    pl = make_plot(qtop)
    savefig(pl, dout*"/qtop.pdf")
end

BDIO_close!(fb)
