###
### "THE BEER-WARE LICENSE":
### Alberto Ramos wrote this file. As long as you retain this 
### notice you can do whatever you want with this stuff. If we meet some 
### day, and you think this stuff is worth it, you can buy me a beer in 
### return. <alberto.ramos@cern.ch>
###
### file:    ym.jl
### created: Wed Oct 27 11:33:34 2021
###                               


using LatticeGPU, TOML, BDIO, ArgParse, Random, CUDA, Pkg, InteractiveUtils, TimerOutputs, Printf, TranscodingStreams, CodecZlib

function parse_commandline()
    s = ArgParseSettings()

    @add_arg_table s begin
        "-i"
        help = "input file"
        required = true
        arg_type = String
    end

    return parse_args(s)
end

function read_options(fname)

    s = TOML.parsefile(fname)

    BDIO_set_user(s["Run"]["user"])
    BDIO_set_host(s["Run"]["host"])

    # Lattice Parameters
    if s["Simulation"]["bcs"] == "PERIODIC"
        BCS = BC_PERIODIC
    elseif s["Simulation"]["bcs"] == "SF"
        BCS = BC_SF_AFWB
    end
    dim = s["Simulation"]["dimensions"]
    lp = SpaceParm{dim}(Tuple(x for x in s["Simulation"]["size"]),
                        Tuple(x for x in s["Simulation"]["blocks"]),
                        BCS, Tuple(x for x in s["Simulation"]["ntwist"]))
    
    # Gauge Parameters
    if s["Simulation"]["group"] == "SU2"
        GRP = SU2
    elseif s["Simulation"]["group"] == "SU3"
        GRP = SU3
    end
    T  = Float64
    if (BCS == BC_SF_AFWB)
        gp = GaugeParm{T}(GRP{T},
                          s["Simulation"]["beta"],
                          s["Simulation"]["c0"],
                          (s["Simulation"]["cG"],s["Simulation"]["cG"]),
                          s["Simulation"]["phiT"], lp.iL[1:end-1])
        phi0 = s["Simulation"]["phi0"]
    else
        gp = GaugeParm{T}(GRP{T},
                          s["Simulation"]["beta"],
                          s["Simulation"]["c0"])
        phi0 = nothing
    end
        
    # HMC parameters
    if s["HMC"]["integrator"] == "OMF4"
        int = omf4(T, s["HMC"]["eps"], s["HMC"]["ns"])
    elseif s["HMC"]["integrator"] == "OMF2"
        int = omf2(T, s["HMC"]["eps"], s["HMC"]["ns"])
    elseif s["HMC"]["integrator"] == "LEAPFROG"
        int = leapfrog(T, s["HMC"]["eps"], s["HMC"]["ns"])
    else
        error("Unknown integrator")
    end
    
    if s["Flow"]["integrator"] == "EULER"
        flwint = wfl_euler(T, s["Flow"]["eps"], 0.0)
    elseif s["Flow"]["integrator"] == "RK2"
        flwint = wfl_rk2(T, s["Flow"]["eps"], 0.0)
    elseif s["Flow"]["integrator"] == "RK3"
        flwint = wfl_rk3(T, s["Flow"]["eps"], 0.0)
    else
        error("Flow integrator not known")
    end
    flw_ns  = s["Flow"]["ns"]
    flw_dtr = s["Flow"]["dtr"]

    # Write basic info to log file
    flog = GzipCompressorStream(open(s["Run"]["name"]*".log.gz", "w+"))
    if GRP == SU2
        println(flog, "\n## HMC SU(2) LatticeGPU simulation")
    else
        println(flog, "\n## HMC SU(3) LatticeGPU simulation")
    end
        
    println(flog, lp)
    println(flog, gp)
    println(flog, int)
    println(flog, flwint)
    
    println(flog, " #  [Enviroment Info]")
    println(flog, "User: ", s["Run"]["user"])
    println(flog, "Host: ", s["Run"]["host"])
    println(flog, " ")

    versioninfo(flog)
    println(flog, " ")

    Pkg.status(io=flog)
#    Pkg.status(mode=Pkg.PKGMODE_MANIFEST, io=flog)
    println(flog, " # END [Enviroment Info]")
    
    fb = BDIO_open(s["Run"]["name"]*".bdio", "w",
                   "HMC simulation with LatticeGPU")
    BDIO_start_record!(fb, BDIO_ASC_GENERIC, 1, true)
    BDIO_write!(fb, read(fname))
    BDIO_write_hash!(fb)

    Random.seed!(CURAND.default_rng(), s["Run"]["seed"])
    Random.seed!(s["Run"]["seed"])

    nthm  = s["HMC"]["nthm"]
    nt    = s["HMC"]["ntraj"]

    return gp, phi0, lp, int, flwint, nthm, nt, flw_ns, flw_dtr, fb, flog, GRP, T, BCS
end

parsed_args = parse_commandline()
infile = parsed_args["i"]

gp, phi0, lp, int, flwint, nthm, nt, flw_ns, flw_dtr, fb, flog, GRP, T, BCS = read_options(infile)

U = vector_field(GRP{T}, lp)
fill!(U, one(GRP{T}))
if phi0 != nothing
    setbndfield(U, phi0, lp)
end

ymws = YMworkspace(GRP, T, lp)

dh = Vector{T}(undef, nthm+nt)
pl = Vector{T}(undef, nthm+nt)
acc = Vector{Bool}(undef, nthm+nt)
if BCS == BC_SF_AFWB
    sfc = Array{T}(undef, nthm+nt, 2)
end

println(flog, " # Thermalization")
HMC!(U,int,lp, gp, ymws, noacc=true)
for i in 1:nthm
    dh[i], acc[i] = HMC!(U,int,lp, gp, ymws)
    pl[i] = plaquette(U,lp, gp, ymws)

    if BCS == BC_PERIODIC
        @printf(flog, "  THM %d/%d:   %s   %6.2e    %20.12e \n",
                i, nthm, acc[i] ? "true " : "false", dh[i], pl[i])
    else
        sfc[i,1], sfc[i,2] = sfcoupling(U, lp, gp, ymws)
        @printf(flog, "  THM %d/%d:   %s   %6.2e    %20.12e     %20.12e %20.12e \n",
                i, nthm, acc[i] ? "true " : "false", dh[i], pl[i],
                sfc[i,1], sfc[i,2])
    end
    if i%10 == 0
        write(flog, TranscodingStreams.TOKEN_END)
        flush(flog)
    end
end


Ecl = Array{T,2}(undef, flw_ns+1, div(nt,flw_dtr))
Epl = Array{T,2}(undef, flw_ns+1, div(nt,flw_dtr))
Qt  = Array{T,2}(undef, flw_ns+1, div(nt,flw_dtr))
println(flog, "\n # Measurements")
kfl = 0
for i in 1:nt
    k = i + nthm
    dh[k], acc[k] = HMC!(U,int,lp, gp, ymws)
    pl[k] = plaquette(U,lp, gp, ymws)

    if BCS == BC_PERIODIC
        @printf(flog, "  MSM %d/%d:   %s   %6.2e    %20.12e \n",
                i, nt, acc[k] ? "true " : "false", dh[k], pl[k])
    else
        sfc[k,1], sfc[k,2] = sfcoupling(U, lp, gp, ymws)
        @printf(flog, "  MSM %d/%d:   %s   %6.2e    %20.12e     %20.12e %20.12e \n",
                i, nt, acc[k] ? "true " : "false", dh[k], pl[k],
                sfc[k,1], sfc[k,2])
    end

    ymws.U1 .= U
    if i%flw_dtr == 0
        global kfl = kfl + 1
        for j in 1:flw_ns
            Epl[j,kfl] = Eoft_plaq(U, gp, lp, ymws)
            Ecl[j,kfl] = Eoft_clover(U, gp, lp, ymws)
            Qt[j,kfl]  = Qtop(U, gp, lp, ymws)

            flw(U, flwint, 1, gp, lp, ymws)
            if (j-1)%div(flw_ns,10) == 0
                t = j*flwint.eps
                @printf(flog, "     FLOW   t=%8.4f:  %20.12e  %20.12e      %10.4e \n",
                        t, t^2*Epl[j,kfl], t^2*Ecl[j,kfl], Qt[j,kfl])
            end
        end
        Epl[end,kfl] = Eoft_plaq(U, gp, lp, ymws)
        Ecl[end,kfl] = Eoft_clover(U, gp, lp, ymws)
        Qt[end,kfl]  = Qtop(U, gp, lp, ymws)
    end
    U .= ymws.U1

    if i%10 == 0
        write(flog, TranscodingStreams.TOKEN_END)
        flush(flog)
    end
end

print_timer(flog)
println(flog, "\n## END HMC LatticeGPU simulation")
close(flog)


# Measurements
#
# uid 2:  dh thermalization
# uid 3:  dh measurements
# uid 4:  pl thermalization
# uid 5:  pl measurements
# uid 8:  dsdeta thermalization
# uid 9:  dsdeta measurements
# uid 10: ddsdnu measurements
# uid 11: ddsdnu measurements
# uid 12: Epl    measurements
# uid 13: Ecl    measurements
# uid 14: Qtop   measurements

function write_record(fb, uid, dt)

    BDIO_start_record!(fb, BDIO_BIN_F64LE, uid,   true)
    BDIO_write!(fb, dt[1:nthm])
    BDIO_write_hash!(fb)
    BDIO_start_record!(fb, BDIO_BIN_F64LE, uid+1, true)
    BDIO_write!(fb, dt[1+nthm:end])
    BDIO_write_hash!(fb)

    return nothing
end

function write_flow_record(fb, uid, dt)

    BDIO_start_record!(fb, BDIO_BIN_F64LE, uid,   true)
    foo = [flwint.eps, convert(T,flw_ns)]
    BDIO_write!(fb, foo)
    for k in 1:div(nt,flw_dtr)
        BDIO_write!(fb, dt[:,k])
    end
    BDIO_write_hash!(fb)

    return nothing
end

write_record(fb, 2, dh)
write_record(fb, 4, pl)
if BCS == BC_SF_AFWB
    write_record(fb, 8, sfc[:,1])
    write_record(fb, 10, sfc[:,2])
end
write_flow_record(fb, 12, Epl)
write_flow_record(fb, 13, Ecl)
write_flow_record(fb, 14, Qt)


BDIO_close!(fb)


